/**
 * ModServ IRC Services
 * Version 2009
 * Copyright (C) 2007, 2008 Christian Jurk <commx@commx.ws>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 3 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <MS_Channels.h>
#include <MS_Main.h>

QHash<QString, MChannel *> channelList;
QHash<char, quint32> cmodeList;

MChannel::MChannel(const QString &strName) {
  m_modeflags = 0;
  m_limit = 0;
  setName(strName);
}

MChannel::~MChannel() {
  QHash<unsigned int, MChanMember *>::iterator mi;
  for(mi = m_memberList.begin(); mi != m_memberList.end(); ++mi) {
    MChanMember *member = mi.value();
    removeMember(member->user());
    delete member;
  }
  
  QList<MChanBan *>::iterator bi;
  for(bi = m_banList.begin(); bi != m_banList.end(); ++bi) {
    MChanBan *ban = *bi;
    delete ban;
  }
  
  m_memberList.clear();
  m_banList.clear();
}

void MChannel::setName(const QString &strName) {
  m_name = strName;
}

void MChannel::setCreationTS(const QDateTime &t_src) {
  m_creationTime = t_src;
}

void MChannel::setModeflags(const quint32 &flags) {
  m_modeflags = flags;
}

void MChannel::setKey(const QString &strKey) {
  m_key = strKey;
}

void MChannel::setLimit(const quint32 &limit) {
  m_limit = limit;
}

void MChannel::setTopic(const QString &strTopic) {
  m_topic = strTopic;
}

const QString &MChannel::name(void) {
  return m_name;
}

const QDateTime &MChannel::creationTS(void) {
  return m_creationTime;
}

const quint32 &MChannel::modeflags(void) {
  return m_modeflags;
}

const QString &MChannel::key(void) {
  return m_key;
}

const quint32 &MChannel::limit(void) {
  return m_limit;
}

const QString &MChannel::topic(void) {
  return m_topic;
}

void MChannel::addMember(MUser *user, const quint32 &flags) {
  if (m_memberList.contains(user->numeric()))
    return; /* user already present on channel */

  m_memberList.insert(user->numeric(), new MChanMember(user, flags));
  user->addChannel(this);
}

void MChannel::removeMember(MUser *user) {
  m_memberList.remove(user->numeric());
  user->removeChannel(this);
  
  if (memberCount() == 0) {
    channelList.remove(name().toLower());
    delete this;
  }
}

MChanMember *MChannel::member(MUser *user) {
  if (m_memberList.contains(user->numeric()))
    return m_memberList.value(user->numeric());
    
  return 0;
}

const QHash<unsigned int, MChanMember *> &MChannel::memberList(void) {
  return m_memberList;
}

void MChannel::clearBanlist(void) {
  m_banList.clear();
}

void MChannel::addBan(const QString &strBanMask, const QString &strBanBy) {
  m_banList.append(new MChanBan(strBanMask, strBanBy));
}

MChanBan *MChannel::findBan(const QString &strBanMask) {
  QList<MChanBan *>::iterator it;
  
  for(it = m_banList.begin(); it != m_banList.end(); ++it) {
    MChanBan *ban = *it;
    if (ban->banmask().toLower() == strBanMask.toLower())
      return ban;
  }
  
  return 0;
}

void MChannel::removeBan(const QString &strBanMask) {
  MChanBan *ban;
  
  if ((ban = findBan(strBanMask)))
    m_banList.removeOne(ban);
}

int MChannel::memberCount(void) {
  return m_memberList.size();
}

int MChannel::banCount(void) {
  return m_banList.size();
}

void MChannel::sendOp(MUser *user, const QString &strSender) {
  MChanMember *cmp = member(user);
  if (!cmp)
    return; /* user is not a member */
    
  quint32 flags = cmp->flags();
  if (FLAG_SET(flags, MChanMember::Flag_ChanOp))
    return; /* user is already opped */

  QString strProtoMsg;
  strProtoMsg = QString("%1 M %2 +o %3")
    .arg(strSender)
    .arg(name())
    .arg(user->numeric64());

  globalMain->socket->send(strProtoMsg);
  FLAG_ADD(flags, MChanMember::Flag_ChanOp);
  cmp->setFlags(flags);
}

void MChannel::sendVoice(MUser *user, const QString &strSender) {
  MChanMember *cmp = member(user);
  if (!cmp)
    return; /* user is not a member */

  quint32 flags = cmp->flags();
  if (FLAG_SET(flags, MChanMember::Flag_Voice))
    return; /* user is already voiced */

  QString strProtoMsg;
  strProtoMsg = QString("%1 M %2 +v %3")
    .arg(strSender)
    .arg(name())
    .arg(user->numeric64());

  globalMain->socket->send(strProtoMsg);
  FLAG_ADD(flags, MChanMember::Flag_Voice);
  cmp->setFlags(flags);
}

void MChannel::sendKickBan(MUser *user, const QString &strSender,
  const QString &strBannedBy, const QString &strBanReason) {
  MChanMember *cmp;
  if (!(cmp = member(user)))
    return; /* user is not a member */

  QString strProtoMsg, strBanMask;
  
  /* build ban mask */
  if (user->isAuthed())
    strBanMask = QString("*!%1@%2.%3")
      .arg(user->ident())
      .arg(user->account())
      .arg(globalMain->config->get("server.hiddenhost", "users.example.org"));
  else
    strBanMask = QString("*!%1@%2")
      .arg(user->ident())
      .arg(user->hostname());

  /* check if the user is opped */
  quint32 flags = cmp->flags();
  if (FLAG_SET(flags, MChanMember::Flag_ChanOp))
    strProtoMsg = QString("%1 M %2 -o+b %3 %4")
      .arg(strSender)
      .arg(name())
      .arg(user->numeric64())
      .arg(strBanMask);
  else
    strProtoMsg = QString("%1 M %2 +b %3")
      .arg(strSender)
      .arg(name())
      .arg(strBanMask);

  globalMain->socket->send(strProtoMsg);
  
  /* send the kick */
  strProtoMsg = QString("%1 K %2 %3 :%4")
    .arg(strSender)
    .arg(name())
    .arg(user->numeric64())
    .arg(strBanReason);
    
  globalMain->socket->send(strProtoMsg);
  addBan(strBanMask, strBannedBy);
  removeMember(user);
}

void MChannel::init(void) {
  if (cmodeList.size() > 0)
    cmodeList.clear();
  
  /* channel mode 'C' (uppercase); Quakenet extension
   * CTCP messages to the channel are blocked.
   */
  cmodeList.insert('C', MChannel::Flag_NoCtcp);
  
  /* channel mode 'c' (lowercase); Quakenet extension
   * Messages with color codes are blocked. Updated versions of the cC patch
   * include blocking all control codes (underline/reverse/bold/...).
   */
  cmodeList.insert('c', MChannel::Flag_NoControlCodes);
  
  /* channel mode 'D' (uppercase); Quakenet extension
   * Users entering the channel are not displayed joining until they speak or
   * get mode changes on the channel.
   * This channel mode is useful for large channels to reduce traffic.
   */
  cmodeList.insert('D', MChannel::Flag_DelayedJoin);
  
  /* channel mode 'd' (lowercase); Quakenet extension
   * When the auditorium mode has been removed and there are still hidden users
   * in the channel, this flag will be set. That flag is set from servers only.
   * The flag gets automatically removed when all users are visible again.
   */
  cmodeList.insert('d', MChannel::Flag_HiddenUsers);
  
  /* channel mode 'i' (lowercase);
   * Only invited users can enter the channel.
   */
  cmodeList.insert('i', MChannel::Flag_InviteOnly);
  
  /* channel mode 'k' (lowercase);
   * Marks the channel password protected. Takes one argument, the channel key.
   */
  cmodeList.insert('k', MChannel::Flag_Key);
  
  /* channel user mode 'l' (lowercase);
   * Defines an channel usercount limit. Only the specified amount of users
   * can enter the channel. Takes one argument, the number of users allowed to
   * enter the channel.
   */
  cmodeList.insert('l', MChannel::Flag_Limit);
  
  /* channel mode 'M' (uppercase); boxenircu extension
   * Channel messages from non-authenticaed users are blocked.
   * Useful for reducing spam.
   */
  cmodeList.insert('M', MChannel::Flag_AuthedMsgOnly);
  
  /* channel mode 'm' (lowercase);
   * Only privileged users can speak in the channel; That includes channel
   * operators, voices and IRC operators.
   */
  cmodeList.insert('m', MChannel::Flag_Moderated);
  
  /* channel mode 'n' (lowercase);
   * No messages targeted to the channel from non-channel members
   */
  cmodeList.insert('n', MChannel::Flag_NoExternalMsg);

  /* channel mode 'N' (uppercase); Quakenet extension
   * Blocks channel wide NOTICE's.
   */
  cmodeList.insert('N', MChannel::Flag_NoNotice);
  
  /* channel mode 'p' (lowercase);
   * Marks the channel being private. This mode flag is known to be outdated.
   * Use flag 's' (secure) instead;
   */
  cmodeList.insert('p', MChannel::Flag_Private);
  
  /* channel mode 'r' (lowercase);
   * Only authorised users can enter the channel.
   */
  cmodeList.insert('r', MChannel::Flag_RegOnly);
  
  /* channel mode 'S' (uppercase); boxenircu extension
   * Strips control codes like bold, underline, colors out of channel messages.
   */
  cmodeList.insert('S', MChannel::Flag_StripCodes);
  
  /* channel mode 's' (lowercase);
   * Marks the channel being secret. It cannot be found using /LIST and so on.
   */
  cmodeList.insert('s', MChannel::Flag_Secure);
  
  /* channel mode 'T' (uppercase); boxenircu extension
   * Multi-target messages (AMSG's) are blocked.
   */
  cmodeList.insert('T', MChannel::Flag_NoMultiTarget);
  
  /* channel mode 't' (lowercase);
   * Only channel operators can change the channel topic.
   */
  cmodeList.insert('t', MChannel::Flag_OnlyOpTopic);
  
  /* channel mode 'u' (lowercase); Quakenet extension
   * Hides quit/part messages from users.
   */
  cmodeList.insert('u', MChannel::Flag_HideQuitPart);
  
  /* channel mode 'z' (lowercase); boxenircu extension
   * This channel flag is just available on servers compiled with SSL support.
   * Only SSL secured users can enter the channel.
   */
  cmodeList.insert('z', MChannel::Flag_SSL);
}

MChannel *MChannel::find(const QString &strName) {
  if (channelList.contains(strName.toLower()))
    return channelList.value(strName.toLower());
    
  return 0;
}

// -- MChanMember --
MChanMember::MChanMember(MUser *user, const quint32 &flags) {
  m_user = user;
  setFlags(flags);
}

void MChanMember::setFlags(const quint32 &flags) {
  m_flags = flags;
}

const quint32 &MChanMember::flags(void) {
  return m_flags;
}

MUser *MChanMember::user(void) {
  return m_user;
}

// -- MChanBan --
MChanBan::MChanBan(const QString &strBanmask, const QString &strBannedby) {
  m_banMask = strBanmask;
  m_banBy = strBannedby;
}

const QString &MChanBan::banmask(void) {
  return m_banMask;
}

const QString &MChanBan::bannedBy(void) {
  return m_banBy;
}
