/**
 * ModServ IRC Services
 * Version 2009
 * Copyright (C) 2007, 2008 Christian Jurk <commx@commx.ws>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 3 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <QTimer>
#include <MS_Networking.h>
#include <MS_Main.h>

MTcpSocket::MTcpSocket(QString strHost, quint16 nPort) {
  m_bytesRx = m_bytesTx = 0.0;
  m_conAttempts = 0;

  connect(this, SIGNAL(connected()), this, SLOT(onSocketConnected()));
  connect(this, SIGNAL(disconnected()), this, SLOT(onSocketDisconnect()));
  connect(this, SIGNAL(error(QAbstractSocket::SocketError)), this, SLOT(onSocketError(QAbstractSocket::SocketError)));
  connect(this, SIGNAL(readyRead()), this, SLOT(onSocketRead()));

  m_remoteHost = strHost;
  m_remotePort = nPort;
  
  connectTo(m_remoteHost, m_remotePort);
}

MTcpSocket::~MTcpSocket() {
  close();
}

void MTcpSocket::connectTo(const QString &strHost, quint16 nPort) {
  if (!strHost.isEmpty() && nPort > 0) {
    close();
    
    if (m_conAttempts >= 10)
      globalMain->log->print(L_FATAL, QString("Unable to connect to remote host, exiting."));

    globalMain->log->print(L_NOTICE, QString("Trying %1 Port %2... ").arg(strHost).arg(nPort));
    connectToHost(strHost, nPort);
    m_conAttempts++;
  }
}

void MTcpSocket::reconnect(void) {
  connectTo(m_remoteHost, m_remotePort);
}

void MTcpSocket::scheduleRetry(void)
{
  if (m_conAttempts < 10)
    globalMain->log->print(L_NOTICE, QString("Reconnect scheduled in 5 seconds..."));

  QTimer::singleShot(5000, this, SLOT(reconnect()));
}

const double &MTcpSocket::getBytesRx(void) {
  return m_bytesRx;
}

const double &MTcpSocket::getBytesTx(void) {
  return m_bytesTx;
}

void MTcpSocket::onSocketConnected(void) {
  globalMain->log->print(L_NOTICE, "Connected.");
  emit readyToAuth();
}

void MTcpSocket::onSocketDisconnect(void) {
  globalMain->log->print(L_NOTICE, "Disconnected.");
}

void MTcpSocket::onSocketError(QAbstractSocket::SocketError sockErr) {
  QString strErrorMsg = QString("Socket error: ");
  
  switch(sockErr) {
    case QAbstractSocket::ConnectionRefusedError:
      strErrorMsg += QString("Connection refused or timed out.");
      break;
    case QAbstractSocket::RemoteHostClosedError:
      strErrorMsg += QString("Remote host closed the connection.");
      break;
    case QAbstractSocket::HostNotFoundError:
      strErrorMsg += QString("Host not found.");
      break;
    case QAbstractSocket::NetworkError:
      strErrorMsg += QString("Network error.");
      break;
    default:
      strErrorMsg += QString("Unknown error occured.");
      break;
  }
  
  globalMain->log->print(L_ERR, strErrorMsg);
  scheduleRetry();
}

void MTcpSocket::onSocketRead(void) {
  static QString strLine;
  char c;

  while(bytesAvailable() > 0) {
    if (!getChar(&c))
      break;

    if (c == '\n' || c == '\0') {
      m_bytesRx += static_cast<double>(strLine.length()) / 1048576.0;
      std::cout << "read(): " << strLine.toStdString() << "\n";
      emit readyToParse(strLine);
      strLine.clear();
      continue;
    }
    
    strLine.append(c);
  }
}

void MTcpSocket::send(const QString &strBuffer, bool appendNl) {
  QString strLine = strBuffer;
  
  if (appendNl)
    strLine.append("\r\n");

  m_bytesTx += static_cast<double>(strLine.length()) / 1024.0 / 1024.0;
  write(strLine.toStdString().c_str(), strLine.length());
  
  std::cout << "send(): " << strLine.toStdString();
}
